(* beta reduction *)
(* $Id: opt.ml,v 1.3 2000/07/18 14:32:44 alamstei Exp $ *)

open Lambda
open List

(* thrown to indicate that a term which was passed to a function with the 
   intention of it being reduced is already in normal form *)
exception CannotReduceAnf;;

(* single step reduction *) 
(* Assume that reduce_leftmost *always* reduces the term, and raise an
   exception if reduce_leftmost is applied to a normal form *)

let rec reduce_leftmost term =

  let rec reduce_first_term termlist =
    match termlist with
      hd::tl -> 
	begin
	  try
	    (reduce_leftmost hd)::tl
	  with
	    CannotReduceAnf -> 
	      hd::(reduce_first_term tl)
	end
    | [] -> raise CannotReduceAnf
  in
  
  let rec reduce_injlist injlist =
    match injlist with
      (a,b,x,m)::tl ->
	begin	
	  try 
	    (a,b,x,(reduce_leftmost m))::tl
	  with
	    CannotReduceAnf -> (a,b,x,m)::(reduce_injlist tl)
        end
    | _ -> []
  in
  
  (* first see if a bsaic function reduction applies *)
  try 
    Stdlib.reduce_opt_step term 
  with
    (* if not, then try any other reduction *)
    Stdlib.Not_reduced ->
      match term with
    	Var(x) -> raise CannotReduceAnf 
      | App(Abs(o, p), n) -> substitute p o n
      | App(m, n) -> 
	  begin
	    try
	      App (reduce_leftmost m, n)
	    with CannotReduceAnf ->
	      App(m, reduce_leftmost n)
	  end
      | Abs(x, m) -> Abs(x, reduce_leftmost m)
      | Let(x,m,n) -> substitute n x m
      | Letrec(x,m,n) -> 
	  begin
	    try
	      let m' = reduce_leftmost m in
	      Letrec(x,m',n)
	    with CannotReduceAnf ->
	      let n' = reduce_leftmost n in
	      Letrec(x,m,n')
	  end
      | Intlit(n) -> raise CannotReduceAnf
      | Boollit(n) -> raise CannotReduceAnf
      | Tuple(termlist) -> Tuple(reduce_first_term termlist)
      | Proj(i,j,m) ->
	  begin
	    try
	      let Tuple(termlist) = m in
	      List.nth termlist (i-1)
	    with
	      Failure("nth") | Match_failure(_,_,_) ->
	    	Proj(i,j,(reduce_leftmost m))
	  end
      | Inj(i,j,n) -> Inj(i,j,(reduce_leftmost n))
      | Case(m,injlist) -> 
	  begin
      	    try
	      let Inj(i,j,n) = m in
      	      let f = fun (a,b,x,p) -> a=i && b=j in
	      let (_,_,x,p) = List.find f injlist in
              substitute p x n
      	    with
	      Not_found | Match_failure (_,_,_) -> 
	    	begin
		  try
	      	    Case((reduce_leftmost m),injlist)
		  with
	      	    CannotReduceAnf -> Case(m,(reduce_injlist injlist))
	    	end
	  end
	    
let rec reduce_to_nf term =
  try 
    reduce_to_nf (reduce_leftmost term) 
  with
    CannotReduceAnf -> term
